(**************************************************************************

   	This file is part of the AtomicDensityMatrix package for Mathematica
    Copyright (C) 2004-2015 Simon M. Rochester <simon@rochesterscientific.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

**************************************************************************)

BeginPackage["AtomicDensityMatrix`Arrow3D`", {"AtomicDensityMatrix`GraphicsUtilities`"}]

(* ::Subsection:: *)
(*Usage messages*)

Axes3D::usage = ""

AutoLengthScale::usage = ""

RenderStyle::usage = ""

Arrow3D::usage = ""

ArrowShape::usage = ""

Width::usage = ""

LabelPosition::usage = ""

HeadAngle::usage = ""

ColorAll::usage = ""

FatHead::usage = ""

FlatHead::usage = ""

Opening::usage = ""

Cone3D::usage = ""

Disk3D::usage = ""

Circle3D::usage = ""

OutlineShadow::usage = ""

WireCylinder::usage = ""

WireFrame::usage = ""

fixArrowheadSpecs::usage = ""

TripleOption::usage = ""

(* ::Subsection:: *)
(*Error messages*)

(* ::Section:: *)
(*Implementation*)

(* ::Subsection:: *)
(*Begin private context*)

Begin["`Private`"]

RotateShape[g_, al_, be_, ga_] := Rotate[#, ga, {0, 0, -1}]&@Rotate[#, be, {-1, 0, 0}]&@Rotate[g, al, {0, 0, -1}]

TranslateShape[g_, {x_, y_, z_}] := Translate[g, {x, y, z}]

(* ::Subsection:: *)
(*Cone3D*)

Options[Cone3D] = {Opening -> Down, PlotPoints -> Automatic, RenderStyle -> Full}

Cone3D[r_, h_, opts : OptionsPattern[]] := Module[{rp3dopts = FilterRules[Join[{opts}, Options[Cone3D]], Options[RevolutionPlot3D]], pp},
    If[OptionValue[RenderStyle] === Full,
    	If[$VersionNumber < 7.,
	        If[OptionValue[Opening] === Down, 
	        	Normal@RevolutionPlot3D[h (1 - t/r), {t, 0, r}, Mesh -> False, Evaluate[Sequence @@ rp3dopts]][[1]],
	            RevolutionPlot3D[h (t/r), {t, 0, r}, Mesh -> False, Evaluate[Sequence @@ rp3dopts]][[1]]
	        ] /. {(Lighting -> _) :> Sequence[], RGBColor[__] :> Sequence[],
	            Specularity[___] :> Sequence[]},
	        If[OptionValue[Opening] === Down, 
	        	{EdgeForm[], Cone[{{0,0,0},{0,0,h}},r]},
	            {EdgeForm[], Cone[{{0,0,h},{0,0,0}},r]}
	        ]
    	],
        pp = OptionValue[PlotPoints] /. Automatic -> 10;
        If[OptionValue[Opening] === Down,
            {Circle3D[{0, 0, 0}, r], Table[Line[{{0, 0, h}, {r Sin[t], r Cos[t], 0}}], {t, 0, 2 Pi - 2 Pi/pp, 2 Pi/pp}]},
            {Circle3D[{0, 0, h}, r], Table[Line[{{0, 0, 0}, {r Sin[t], r Cos[t], h}}], {t, 0, 2 Pi - 2 Pi/pp, 2 Pi/pp}]}
        ]
    ]
]

(* ::Subsection:: *)
(*Disk3D*)

Disk3D[{x_, y_, z_}, r_] := Translate[RevolutionPlot3D[0, {t, 0, r}, Mesh -> False][[1]], {x, y, z}]

(* ::Subsection:: *)
(*WireCylinder*)

WireCylinder[{p1 : {x1_, y1_, z1_}, p2 : {x2_, y2_, z2_}}, opts___Rule] := WireCylinder[{p1, p2}, 1, opts]

WireCylinder[{p1 : {x1_, y1_, z1_}, p2 : {x2_, y2_, z2_}}, r_, opts___Rule] :=
    Translate[Rotate[WireCylinder[Sqrt[Total[(p2 - p1)^2]], r, opts], {{0, 0, 1}, (p2 - p1)/Sqrt[Total[(p2 - p1)^2]]}], p1]

WireCylinder[h_, r_, opts___Rule] := RevolutionPlot3D[{r, t}, {t, 0, h}, Mesh -> False][[1]] /. Polygon -> Line

(* ::Subsection:: *)
(*OutlineShadow*)

Options[OutlineShadow] = {BaseStyle -> Directive[White, AbsoluteThickness[4]]}

OutlineShadow[vp : {_, _, _}, d_, OptionsPattern[]][g_] := {
    g, Dashing[{}], OptionValue[BaseStyle],
    GeometricTransformation[g /. {GrayLevel[_] -> {}, RGBColor[__] -> {}, Thickness -> {},
        AbsoluteThickness -> {}, Text[__] -> {}, Polygon -> Line, Cylinder -> WireCylinder},
        {(1 + d)IdentityMatrix[3], -d vp}
    ]
}

(* ::Subsection:: *)
(*Arrow3D*)

Options[Arrow3D] = {ArrowShape -> {.1, .03, .01}, Width -> .01, Label -> None,
    LabelPosition -> {1, 10, \[Pi]/2}, LabelStyle -> {}, RenderStyle -> Full,
    Arrowheads -> Automatic, BaseStyle -> {}, PlotPoints -> 5, Scale -> Automatic};

Arrow3D[opts___Rule] := Arrow3D[{{0, 0, 0}, {1, 0, 0}}, opts]

Arrow3D[p : {x_, y_, z_}, opts___Rule] := Arrow3D[{{0, 0, 0}, p}, opts]

Arrow3D[{p1 : {x1_, y1_, z1_}, p2 : {x2_, y2_, z2_}}, opts___Rule] /; Norm[p1 - p2] == 0. := {}

Arrow3D[{p1 : {x1_, y1_, z1_}, p2 : {x2_, y2_, z2_}}, opts___Rule] :=
    Translate[Rotate[Arrow3D[Sqrt[Total[(p2 - p1)^2]], opts], {{0, 0, 1}, (p2 - p1)/Sqrt[Total[(p2 - p1)^2]]}], p1]

FatHead[length_, width_, inset_, opts___Rule] := TranslateZ[-inset - length]@{Cone3D[width, inset, Opening -> Down, opts], Cone3D[width, inset + length, Opening -> Down, opts]}

FlatHead[length_, width_, inset_, opts___Rule] := TranslateZ[-length]@Polygon[{{-width, 0, -inset}, {0, 0, length}, {width, 0, -inset}, {0, 0, 0}, {-width, 0, -inset}}]

fixArrowheadSpec[s_, i_, n_, default_, scale_] := Module[{spec = s, defs},
    If[Head[spec] =!= List, spec = {spec}];
    defs = {1, If[n == 1, 1, (i - 1)/(n - 1)], default, {0, 0}};
    spec = Join[spec, Drop[defs, Length[spec]]];
    spec = MapThread[(#1 /. Automatic -> #2)&, {spec, defs}];
    {spec[[2]], scale spec[[1]]spec[[3]], spec[[4]]}
]

fixArrowheadSpecs[s_, default_, scale_, length_] := Module[{heads = s},
    heads = If[Head[heads] =!= List, {heads}, heads];
    heads = Sort[
        MapIndexed[fixArrowheadSpec[#1, #2[[1]], Length[heads], default, scale]&, heads] /.
            Automatic -> length,
        OrderedQ[{#1[[2]], #2[[2]]}]&
    ]
]

Arrow3D[length_, opts : OptionsPattern[]] :=
    Module[{scale, width, labpos, labstyle, heads = OptionValue[Arrowheads],
        rs = OptionValue[RenderStyle], basestyle},
        {width, scale, labpos, labstyle, basestyle} =
            {Scale Width, Scale, LabelPosition, LabelStyle, BaseStyle} /.
                {opts} /. Options[Arrow3D] /. Automatic -> length;
        (*Massage arrowheads specifications*)
        heads = fixArrowheadSpecs[OptionValue[Arrowheads], OptionValue[ArrowShape], OptionValue[Scale], length];
        (*Graphics statement*)
        {If[Head[basestyle] === List, Sequence @@ basestyle, basestyle],
        (*shorten arrow shaft if there are arrowheads at head or tail*)
            Switch[rs, Full | FlatHead, {EdgeForm[], Cylinder[#, width]}, Flat, {Line[#]}]&@{{0, 0, If[Length[heads] > 0 && heads[[1, 1]] == 0 && heads[[1, 2, 1]] < 0, -heads[[1, 2, 1]], 0]}, {0, 0, length - If[Length[heads] > 0 && heads[[-1, 1]] == 1 && heads[[-1, 2, 1]] > 0, heads[[-1, 2, 1]], 0]}},
            TranslateZ[length #[[1]]]@RotateZ[#[[3, 2]]]@RotateX[#[[3, 1]]]@Switch[rs, Full, FatHead, Flat | FlatHead, FlatHead][Sequence @@ #[[2]], PlotPoints -> OptionValue[PlotPoints]]& /@ heads,
            If[OptionValue[Label] =!= None, RotateZ[labpos[[3]]]@Style[Text[OptionValue[Label], {width labpos[[2]], 0, length labpos[[1]]}, Sequence @@ If[Length[labpos] > 3, Drop[labpos, 3], {}]], labstyle], {}]
        }
    ]

(* ::Subsection:: *)
(*Axes3D*)

(* ::Text:: *)
(*In order x, y, z axis*)

(*Old default options*)
(*Options[Axes3D]={Label->{"x","y","z"},LabelPosition -> {{1.04, 0, 0}, {1.1, 0, 0}, {1.05, 0, 0}},
	AutoLengthScale->1.3,ArrowShape->{.1,.03,.01},Width->.01,Scale->Automatic,
	LabelStyle->Directive[FontFamily->"Times New Roman",FontSize->14,FontSlant->"Italic",FontWeight->"Bold",Black],
	RenderStyle->Full,Arrowheads->Automatic,BaseStyle->{},Range->{{-1,1},{-1,1},{-1,1}}};
*)

Options[Axes3D] = {Label -> {"x", "y", "z"}, LabelPosition -> {{1.04, 0, 0}, {1.1, 0, 0}, {1.07, 0, 0}},
    AutoLengthScale -> 1.5, ArrowShape -> .9{.1, .03, .01}, Width -> .01, Scale -> Automatic,
    LabelStyle -> Directive[FontFamily -> "Times New Roman", FontSize -> 14, FontSlant -> "Italic", FontWeight -> "Bold", Black],
    RenderStyle -> Full, Arrowheads -> Automatic, BaseStyle -> Gray, Range -> {{-1, 1}, {-1, 1}, {-1, 1}}};

TripleOption[expr_] := If[Head[expr] =!= List, Table[expr, {3}], expr];

Axes3D[l : {lx_, ly_, lz_}, opts : OptionsPattern[]] := Module[{labpos, lab, labelstyle, type, heads, basestyle},
    {labpos, lab, labelstyle, type, heads, basestyle} = {LabelPosition, Label, LabelStyle, RenderStyle, Arrowheads, BaseStyle} /. {opts} /. Options[Axes3D];
    {lab, type, heads} = TripleOption /@ {lab, type, heads};
    If[!Length@Dimensions@labpos > 1, labpos = Table[labpos, {3}]];
    MapAt[RotateX[\[Pi]/2], #, 1]&@MapThread[
        Arrow3D[#1, Label -> #2, LabelPosition -> #3, RenderStyle -> #4, Arrowheads -> #5,
            Sequence @@ FilterRules[{opts}, Options[Arrow3D]],
            Sequence @@ FilterRules[Options[Axes3D], Options[Arrow3D]]]&,
        {OptionValue[Range]{{{lx, 0, 0}, {lx, 0, 0}}, {{0, ly, 0}, {0, ly, 0}}, {{0, 0, lz}, {0, 0, lz}}}, lab, labpos, type, heads}
    ]
]

Axes3D[l_?NumericQ, opts___Rule] := Axes3D[{l, l, l}, opts]

Axes3D[opts___Rule] := Axes3D[1, opts]

Axes3D[g_Graphics3D, l : {lx_, ly_, lz_}, opts___Rule] :=
    Show[g, Graphics3D @ Axes3D[{lx, ly, lz}, opts]]

Axes3D[g_Graphics3D, l_?NumericQ, opts___Rule] := Axes3D[g, {l, l, l}, opts]

Axes3D[g_Graphics3D, opts___Rule] := Module[{length, als},
    als = AutoLengthScale /. {opts} /. Options[Axes3D];
    length = als Max@Abs[PlotRange /. AbsoluteOptions[g, PlotRange]];
    Axes3D[g, length, opts]
]

(* ::Subsection:: *)
(*Circle3D*)

Options[Circle3D] = {Label -> None, LabelPosition -> {0, 1.2, 0}, LabelStyle -> {}, RenderStyle -> Flat,
    Width -> .01, Arrowheads -> {}, ArrowShape -> {.1, .03, .01}, Scale -> Automatic,
    PlotPoints -> {Automatic, Automatic}
}

Circle3D[r_, opts___Rule] := Circle3D[{0, 0, 0}, r, {0, 2\[Pi]}, opts]

Circle3D[p : {x_, y_, z_}, r_, opts___Rule] := Circle3D[p, r, {0, 2\[Pi]}, opts]

Circle3D[p : {x_, y_, z_}, r_, {t1_, t2_}, opts : OptionsPattern[]] := Module[{lp = OptionValue[LabelPosition], w, rs = OptionValue[RenderStyle], ppopts, ah, t1a, t2a, ppt = OptionValue[PlotPoints]},
    {w} = {Scale Width} /. {opts} /. Options[Circle3D] /. Automatic -> r;
    ah = fixArrowheadSpecs[OptionValue[Arrowheads], OptionValue[ArrowShape], OptionValue[Scale], Sign[t2 - t1]r];
    ppopts = DeleteCases[FilterRules[Join[{opts}, Options[Circle3D]], Options[ParametricPlot3D]], PlotPoints -> _];
    {t1a, t2a} = {t1, t2} - {If[Length[ah] > 0 && ah[[1, 1]] == 0 && Sign[t2 - t1]ah[[1, 2, 1]] < 0, ah[[1, 2, 1]]/r, 0], If[Length[ah] > 0 && ah[[-1, 1]] == 1 && Sign[t2 - t1]ah[[-1, 2, 1]] > 0, ah[[-1, 2, 1]]/r, 0]};
    Translate[{
        Switch[rs,
            Full,
            ParametricPlot3D[{(r + w Cos[v]) Cos[u], (r + w Cos[v]) Sin[u], w Sin[v]}, {u, t1a, t2a}, {v, 0, 2\[Pi]}, Mesh -> False, PlotPoints -> ppt[[1]], Evaluate[Sequence @@ ppopts]][[1]],
            Flat, GetLine@ParametricPlot3D[r {Cos[t], Sin[t], 0}, {t, t1a, t2a}, PlotPoints -> ppt[[1]], Evaluate@ppopts]
        ],
        RotateZ[t1 + #[[1]](t2 - t1)]@TranslateX[r]@RotateX[-\[Pi]/2]@RotateY[#[[2, 1]]/(2r)]@Switch[rs, Full, FatHead, Flat, FlatHead][Sequence @@ #[[2]], PlotPoints -> ppt[[2]]]& /@ ah,
        If[OptionValue[Label] =!= None, RotateZ[(t1 + t2)/2 + lp[[1]](t2 - t1)/2]@Text[OptionValue[Label], {r lp[[2]], 0, r lp[[3]]}, Sequence @@ If[Length[lp] > 3, Drop[lp, 3], {}], If[Head[#] === List, Sequence @@ #, #]&@OptionValue[LabelStyle]], {}]
    },
        p
    ]
]

(* ::Subsection:: *)
(*End the private context*)

End[]

(* ::Section:: *)
(*Epilog*)

(* ::Text:: *)
(*This section protects exported symbols and ends the package.*)

(* ::Subsection:: *)
(*Protect exported symbols*)

(*Protect[Evaluate[$Context<>"*"]]*)

(* ::Subsection:: *)
(*End the package context*)

EndPackage[]
